'use strict';

var fs = require('fs');
var path = require('path');
var findBabelConfig = require('find-babel-config');
var glob = require('glob');
var requireResolve = require('resolve');

var defaultExtensions = ['.js', '.jsx', '.es', '.es6'];

function isRegExp(string) {
  return string.startsWith('^') || string.endsWith('$');
}

function normalizeCwd(opts, file) {
  if (opts.cwd === 'babelrc') {
    var startPath = file.opts.filename === 'unknown' ? './' : file.opts.filename;

    var _findBabelConfig$sync = findBabelConfig.sync(startPath),
        babelPath = _findBabelConfig$sync.file;

    opts.cwd = babelPath ? path.dirname(babelPath) : null;
  }

  if (!opts.cwd) {
    opts.cwd = process.cwd();
  }
}

function normalizeRoot(opts) {
  if (opts.root) {
    if (!Array.isArray(opts.root)) {
      opts.root = [opts.root];
    }

    opts.root = opts.root.map(function (dirPath) {
      return path.resolve(opts.cwd, dirPath);
    }).reduce(function (resolvedDirs, absDirPath) {
      if (glob.hasMagic(absDirPath)) {
        var roots = glob.sync(absDirPath).filter(function (path$$1) {
          return fs.lstatSync(path$$1).isDirectory();
        });

        return [].concat(resolvedDirs, roots);
      }

      return [].concat(resolvedDirs, [absDirPath]);
    }, []);
  } else {
    opts.root = [];
  }
}

function getAliasPair(key, value) {
  var parts = value.split('\\\\');

  function substitute(execResult) {
    return parts.map(function (part) {
      return part.replace(/\\\d+/g, function (number) {
        return execResult[number.slice(1)] || '';
      });
    }).join('\\');
  }

  return [new RegExp(key), substitute];
}

function normalizeAlias(opts) {
  if (opts.alias) {
    var alias = opts.alias;

    var aliasKeys = Object.keys(alias);

    var nonRegExpAliases = aliasKeys.filter(function (key) {
      return !isRegExp(key);
    }).map(function (key) {
      return getAliasPair(`^${key}((?:/|).*)`, `${alias[key]}\\1`);
    });

    var regExpAliases = aliasKeys.filter(isRegExp).map(function (key) {
      return getAliasPair(key, alias[key]);
    });

    opts.alias = [].concat(nonRegExpAliases, regExpAliases);
  } else {
    opts.alias = [];
  }
}

function normalizeOptions(opts, file) {
  normalizeCwd(opts, file);
  normalizeRoot(opts);
  normalizeAlias(opts);

  if (!opts.extensions) {
    opts.extensions = defaultExtensions;
  }
}

function mapToRelative(cwd, currentFile, module) {
  var from = path.dirname(currentFile);
  var to = path.normalize(module);

  from = path.resolve(cwd, from);
  to = path.resolve(cwd, to);

  var moduleMapped = path.relative(from, to);
  return toPosixPath(moduleMapped);
}

function findPathInRoots(sourcePath, _ref) {
  var extensions = _ref.extensions,
      root = _ref.root;

  var resolvedSourceFile = void 0;

  root.some(function (basedir) {
    try {
      // Check if the file exists (will throw if not)
      resolvedSourceFile = requireResolve.sync(`./${sourcePath}`, {
        basedir,
        extensions
      });
      return true;
    } catch (e) {
      return false;
    }
  });

  return resolvedSourceFile;
}

function getRealPathFromRootConfig(sourcePath, currentFile, opts) {
  var absFileInRoot = findPathInRoots(sourcePath, opts);

  if (!absFileInRoot) {
    return null;
  }

  var realSourceFileExtension = path.extname(absFileInRoot);
  var sourceFileExtension = path.extname(sourcePath);

  // Map the source and keep its extension if the import/require had one
  var ext = realSourceFileExtension === sourceFileExtension ? realSourceFileExtension : '';
  return toLocalPath(toPosixPath(replaceExtension(mapToRelative(opts.cwd, currentFile, absFileInRoot), ext)));
}

function getRealPathFromAliasConfig(sourcePath, currentFile, opts) {
  var aliasedSourceFile = void 0;

  opts.alias.find(function (_ref2) {
    var regExp = _ref2[0],
        substitute = _ref2[1];

    var execResult = regExp.exec(sourcePath);

    if (execResult === null) {
      return false;
    }

    aliasedSourceFile = substitute(execResult);
    return true;
  });

  if (!aliasedSourceFile) {
    return null;
  }

  if (aliasedSourceFile[0] === '.') {
    return aliasedSourceFile;
  }

  var realPathFromRoot = getRealPathFromRootConfig(aliasedSourceFile, currentFile, opts);
  if (realPathFromRoot) {
    return realPathFromRoot;
  }

  return aliasedSourceFile;
}

function getRealPath(sourcePath, _ref3) {
  var file = _ref3.file,
      opts = _ref3.opts;

  if (sourcePath[0] === '.') {
    return sourcePath;
  }

  // File param is a relative path from the environment current working directory
  // (not from cwd param)
  var currentFile = path.resolve(file.opts.filename);

  var sourceFileFromRoot = getRealPathFromRootConfig(sourcePath, currentFile, opts);
  if (sourceFileFromRoot) {
    return sourceFileFromRoot;
  }

  var sourceFileFromAlias = getRealPathFromAliasConfig(sourcePath, currentFile, opts);
  if (sourceFileFromAlias) {
    return sourceFileFromAlias;
  }

  return sourcePath;
}

function toPosixPath(modulePath) {
  return modulePath.replace(/\\/g, '/');
}

function toLocalPath(path$$1) {
  return path$$1.replace(/\/index$/, '') // remove trailing /index
  .replace(/^(?!\.)/, './'); // insert `./` to make it a local path
}

function replaceExtension(path$$1, ext) {
  var filename = path.basename(path$$1, path.extname(path$$1)) + ext;
  return path.join(path.dirname(path$$1), filename);
}

function matchesPattern(types, calleePath, pattern) {
  var node = calleePath.node;


  if (types.isMemberExpression(node)) {
    return calleePath.matchesPattern(pattern);
  }

  if (!types.isIdentifier(node) || pattern.includes('.')) {
    return false;
  }

  var name = pattern.split('.')[0];

  return node.name === name;
}

function mapPathString(types, nodePath, state) {
  if (!types.isStringLiteral(nodePath)) {
    return;
  }

  var modulePath = getRealPath(nodePath.node.value, state);
  if (modulePath) {
    nodePath.replaceWith(types.stringLiteral(modulePath));
  }
}

var patterns = ['require', 'require.resolve', 'System.import', 'jest.genMockFromModule', 'jest.mock', 'jest.unmock', 'jest.doMock', 'jest.dontMock'];

function transformCall(types, nodePath, state) {
  var calleePath = nodePath.get('callee');

  if (!patterns.some(function (pattern) {
    return matchesPattern(types, calleePath, pattern);
  })) {
    return;
  }

  mapPathString(types, nodePath.get('arguments.0'), state);
}

function transformImport(types, nodePath, state) {
  mapPathString(types, nodePath.get('source'), state);
}

var index = (function (_ref) {
  var types = _ref.types;

  var importVisitors = {
    CallExpression(nodePath, state) {
      transformCall(types, nodePath, state);
    },
    ImportDeclaration(nodePath, state) {
      transformImport(types, nodePath, state);
    },
    ExportDeclaration(nodePath, state) {
      transformImport(types, nodePath, state);
    }
  };

  return {
    pre(file) {
      normalizeOptions(this.opts, file);
    },

    visitor: {
      Program: {
        exit(programPath, state) {
          programPath.traverse(importVisitors, state);
        }
      }
    }
  };
});

module.exports = index;
